/* life3dviewer.c
 * Implementation of display code for 3D life
 * Operating Systems - 304-427A, Fall 1998
 *
 * Nadim Abdo
 */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

/*motif headers*/
#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <Xm/ToggleB.h>
#include <Xm/CascadeB.h>
#include <Xm/Frame.h>
#include <X11/keysym.h>
/* Motif OpenGL drawing area widget */
#include <X11/GLw/GLwMDrawA.h>
/* openGL headers*/
#include <GL/gl.h>
#include <GL/glx.h>
#include <GL/glu.h>	      

#include "life3dviewer.h"

#define CUBE_LIST      1
#define N CELLS_PER_EDGE

/*cell scale factor, controls gap size between cells*/
#define CELL_SCALE 0.7
/*size of cell/size of parent cube*/
#define SCALE_F CELL_SCALE/CELLS_PER_EDGE
/*cube width*/
#define CUBE_WIDTH 2.0
/*1/2 * width of cells*/
#define   HALF_CELL_WIDTH CUBE_WIDTH/CELLS_PER_EDGE*0.5
#define NUM_FACES 6

/*appearance settings*/
#define SET_COLOR_LIVE glColor3f(1.0,0.0,0.0)
#define SET_COLOR_MAIN_CUBE glColor3f(0.0,0.1,0.6)
#define SET_COLOR_DEAD glColor3f(0.1,0.1,0.1)

#define FACE_1 0
#define FACE_2 1
#define FACE_3 2
#define FACE_4 3
#define FACE_5 4
#define FACE_6 5

#define TRUE 1
#define FALSE 0


static int  dblBuf[] =  {GLX_DOUBLEBUFFER, GLX_RGBA,
                         GLX_RED_SIZE, 1,
                         GLX_BLUE_SIZE, 1,
                         GLX_GREEN_SIZE, 1,
                         GLX_DEPTH_SIZE, 1,
                         None};
static int* snglBuf = &(dblBuf[1]);

GLfloat lightZeroPosition[] = {3.0, 3.0, 1.0, 1.0};
GLfloat lightZeroAmbient[] =  {0.6, 0.6, 0.6, 1.0};

GLfloat lightZeroColor[] =    {0.4, 0.4, 1.0, 1.0}; /* blue-tinted */


Display        *dpy;
Window          win;
GLfloat         g_xAngle = 42.0, g_yAngle = 82.0, g_zAngle = 112.0;

GLboolean    doubleBuffer = GL_TRUE;
GLboolean    bInMotion    = GL_FALSE;
GLboolean    made_current = GL_FALSE;
Boolean      bLibInitDone  =False;
Boolean      g_bAutoRotate =True;

XtAppContext app;
XtWorkProcId workIdAnim = 0;
Widget       toplevel, mainw, menubar, menupane, btn, cascade, frame, glxarea;
GLXContext   cx;
XVisualInfo *vi;
Arg          menuPaneArgs[1], args[1];

/*last mouse pos*/
int g_iLastX,g_iLastY;

/*local copies of live, dead cell info, 6 NxN arrays*/
int g_Faces[NUM_FACES][N][N];

static String   fallbackResources[] = {
#ifdef IRIX_5_2_or_higher
  "*sgiMode: true",		/* SGI looks cool*/
  "*useSchemes: all",
#endif
  "*title: Life3D Viewer - OS A3 (v1.0)",
  "*glxarea*width: 600", "*glxarea*height: 300", NULL
};

void test_RandomCells()
{
  int n,i,j;
  
  for(n=0;n<NUM_FACES;n++)
    for(i=0;i<N;i++)
      for(j=0;j<N;j++)
        g_Faces[n][i][j]=rand()%2;
}

void
fatalError(char *message)
{
    fprintf(stderr, "life3dlib: %s\n", message);
    exit(1);
}

void makeCubeList()
{
  glNewList(CUBE_LIST,GL_COMPILE);

  glShadeModel(GL_FLAT);

  /* Draw the sides of the cube */
  glBegin(GL_QUAD_STRIP);
  glVertex3d(1.0, 1.0, -1.0);
  glVertex3d(1.0, -1.0, -1.0);
  glVertex3d(-1.0, 1.0, -1.0);
  glVertex3d(-1.0, -1.0, -1.0);
  glVertex3d(-1.0, 1.0, 1.0);
  glVertex3d(-1.0, -1.0, 1.0);
  glVertex3d(1.0, 1.0, 1.0);
  glVertex3d(1.0, -1.0, 1.0);
  glVertex3d(1.0, 1.0, -1.0);
  glVertex3d(1.0, -1.0, -1.0);
  glEnd();
                
  /* Draw the top and bottom of the cube
     Culling enabled so faces have opposite
     vertex orderings
  */
  glBegin(GL_QUADS);
  glVertex3d(-1.0, -1.0, -1.0);
  glVertex3d(1.0, -1.0, -1.0);
  glVertex3d(1.0, -1.0, 1.0);
  glVertex3d(-1.0, -1.0, 1.0);
  
  glVertex3d(-1.0, 1.0, 1.0);
  glVertex3d(1.0, 1.0, 1.0);
  glVertex3d(1.0, 1.0, -1.0);
  glVertex3d(-1.0, 1.0, -1.0);
  
  glEnd();
  glEndList();
}

#define GRID_LIST 2
void make2DGridList()
{
  glNewList(GRID_LIST, GL_COMPILE);

  glBegin(GL_QUADS);
  glVertex2f(1.0,1.0);
  glVertex2f(0.0,1.0);
  glVertex2f(0.0,0.0);
  glVertex2f(1.0,0.0);

  glEnd();

  glEndList();
}


void
draw(Widget w)
{
  if(doubleBuffer)
    glXSwapBuffers(dpy, XtWindow(w)); /* buffer swap does implicit glFlush */
  else
    glFlush(); /* explicit flush for single buffered case */
  if(!glXIsDirect(dpy,cx))
    glFinish();   /* avoid queing latency*/

#ifdef DEBUG
  /* for help debugging, report any OpenGL errors that occur per frame */
  GLenum error;
  while((error = glGetError()) != GL_NO_ERROR)
    fprintf(stderr, "GL error: %s\n", gluErrorString(error));
#endif
}

void Update3dViewPort()
{
  int i,j;
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-1.0, 1.0, -1.0, 1.0, 1.0, 10.0);

  glMatrixMode(GL_MODELVIEW);
  /* reset modelview matrix to the identity matrix */
  glLoadIdentity();
  /* rotate by X, Y, and Z angles */
  glTranslatef(0.0, 0.0, -3.5);
  glRotatef(g_xAngle, 0.1, 0.0, 0.0);
  glRotatef(g_yAngle, 0.0, 0.1, 0.0);
  glRotatef(g_zAngle, 0.0, 0.0, 1.0);

  glPushMatrix();
  /*draw the center cube*/
  glLightfv(GL_LIGHT0, GL_POSITION, lightZeroPosition);
  SET_COLOR_MAIN_CUBE;

  glCallList(CUBE_LIST);

  /*gratuitous use of cut'n'paste follows*/  
  for(i=0;i<CELLS_PER_EDGE;i++)
    {
      for(j=0;j<CELLS_PER_EDGE;j++)
        {
          glPopMatrix();
          glPushMatrix();
          
          glTranslatef(-1.0+HALF_CELL_WIDTH+HALF_CELL_WIDTH*2*i,
                       1.0-HALF_CELL_WIDTH-HALF_CELL_WIDTH*2*j,
                       1.0+HALF_CELL_WIDTH);
          glLightfv(GL_LIGHT0, GL_POSITION, lightZeroPosition);
          glScalef(SCALE_F,SCALE_F,SCALE_F);
          if(g_Faces[FACE_1][i][j])
            {
              SET_COLOR_LIVE;
            }
          else
            {
              /*dead*/
              SET_COLOR_DEAD;
            }
          glCallList(CUBE_LIST);
        }
    }
  
  for(i=0;i<CELLS_PER_EDGE;i++)
    {
      for(j=0;j<CELLS_PER_EDGE;j++)
        {
          glPopMatrix();
          glPushMatrix();
          
          glTranslatef(1.0+HALF_CELL_WIDTH,
                       1.0-HALF_CELL_WIDTH-HALF_CELL_WIDTH*2*j,
                       1.0-HALF_CELL_WIDTH-HALF_CELL_WIDTH*2*i);
          glScalef(SCALE_F,SCALE_F,SCALE_F);
          glLightfv(GL_LIGHT0, GL_POSITION, lightZeroPosition);
          if(g_Faces[FACE_2][i][j])
            {
              SET_COLOR_LIVE;
            }
          else
            {
              /*dead*/
              SET_COLOR_DEAD;
            }
          glCallList(CUBE_LIST);
        }
    }
   
    for(i=0;i<CELLS_PER_EDGE;i++)
      {
        for(j=0;j<CELLS_PER_EDGE;j++)
          {
            glPopMatrix();
            glPushMatrix();
            
            glTranslatef(1.0-HALF_CELL_WIDTH-HALF_CELL_WIDTH*2*i,
                         1.0-HALF_CELL_WIDTH-HALF_CELL_WIDTH*2*j,
                         -1.0-HALF_CELL_WIDTH);
            glScalef(SCALE_F,SCALE_F,SCALE_F);
            glLightfv(GL_LIGHT0, GL_POSITION, lightZeroPosition);
            if(g_Faces[FACE_3][i][j])
              {
                SET_COLOR_LIVE;
              }
            else
              {
                /*dead*/
                SET_COLOR_DEAD;
              }
            glCallList(CUBE_LIST);
          }
      }

    for(i=0;i<CELLS_PER_EDGE;i++)
      {
        for(j=0;j<CELLS_PER_EDGE;j++)
          {
            glPopMatrix();
            glPushMatrix();
            
            glTranslatef(-1.0-HALF_CELL_WIDTH,
                         1.0-HALF_CELL_WIDTH-HALF_CELL_WIDTH*2*j,
                         -1.0+HALF_CELL_WIDTH+HALF_CELL_WIDTH*2*i);
            glScalef(SCALE_F,SCALE_F,SCALE_F);
            glLightfv(GL_LIGHT0, GL_POSITION, lightZeroPosition);
            if(g_Faces[FACE_4][i][j])
              {
                SET_COLOR_LIVE;
              }
            else
              {
                /*dead*/
                SET_COLOR_DEAD;
              }

            glCallList(CUBE_LIST);
          }
      }

    for(i=0;i<CELLS_PER_EDGE;i++)
      {
        for(j=0;j<CELLS_PER_EDGE;j++)
          {
            glPopMatrix();
            glPushMatrix();
            
            glTranslatef(-1.0+HALF_CELL_WIDTH+HALF_CELL_WIDTH*2*i,
                         1.0+HALF_CELL_WIDTH,
                         1.0-HALF_CELL_WIDTH-HALF_CELL_WIDTH*2*j);
            glScalef(SCALE_F,SCALE_F,SCALE_F);
            glLightfv(GL_LIGHT0, GL_POSITION, lightZeroPosition);
            if(g_Faces[FACE_5][i][j])
              {
                SET_COLOR_LIVE;
              }
            else
              {
                /*dead*/
                SET_COLOR_DEAD;
              }
            glCallList(CUBE_LIST);
          }
      }
    
    for(i=0;i<CELLS_PER_EDGE;i++)
      {
        for(j=0;j<CELLS_PER_EDGE;j++)
          {
            glPopMatrix();
            glPushMatrix();
            
            glTranslatef(-1.0+HALF_CELL_WIDTH+HALF_CELL_WIDTH*2*i,
                         -1.0-HALF_CELL_WIDTH,
                         1.0-HALF_CELL_WIDTH-HALF_CELL_WIDTH*2*j);
            glScalef(SCALE_F,SCALE_F,SCALE_F);
            glLightfv(GL_LIGHT0, GL_POSITION, lightZeroPosition);
            if(g_Faces[FACE_6][i][j])
              {
                SET_COLOR_LIVE;
              }
            else
              {
                /*dead*/
                SET_COLOR_DEAD;
              }

            glCallList(CUBE_LIST);
          }
      }

    

    glPopMatrix();
}

void drawFaceCells(int faceNum)
{
  int i,j;
  assert(faceNum >= 0 && faceNum < 6);

  
  for(i=0;i<CELLS_PER_EDGE;i++)
    {
      for(j=0;j<CELLS_PER_EDGE;j++)
        {
          glPushMatrix();
          glTranslatef(1.0/N*i,
                       1.0-1.0/N*j,
                       -0.5);

          glScalef(0.8/N,0.8/N,0.8/N);
          if(g_Faces[faceNum][i][j])
            glColor3f(1.0,0.0,0.0);
          else
            glColor3f(0.0,0.2,0.4);
          glCallList(GRID_LIST);
          glPopMatrix();
        }
    }
}

GLfloat facePositions[] = {0.025,0.6,
                           0.35,0.6,
                           0.675,0.6,
                           0.025,0.2,
                           0.35,0.2,
                           0.675,0.2
};

void Update2dViewPort()
{
  int i;
  glMatrixMode(GL_PROJECTION);

  // Set the projection matrix to be the identity matrix
  glLoadIdentity();
  // Define the dimensions of the Orthographic Viewing Volume
  glOrtho(-0.0, 1.0, -0.0, 1.0, -8.0, 8.0);

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  /*  glLightfv(GL_LIGHT0, GL_POSITION, lightZeroPosition);*/

  /*draw each of the 6 cells*/
  for(i=0;i<12;i+=2)
    {
      glLoadIdentity();
      glTranslatef(facePositions[i],
                   facePositions[i+1],0.0);
      glScalef(0.3,0.3,0.3);
      /*draw the underlying cells*/
      drawFaceCells(i/2); /*i/2 = face num*/
    }
  
  glPopMatrix();
}


void UpdateView()
{
  int i,j;
  Dimension       width, height;
  
  XtVaGetValues(glxarea, XmNwidth, &width, XmNheight, &height, NULL);
  glViewport((GLint) width/2, 0, (GLint) width/2, (GLint) height);  
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  Update3dViewPort();

  glViewport(0, 0, (GLint) width/2, (GLint) height);  
  
  Update2dViewPort();
  
  draw(glxarea);
}

Boolean animate(XtPointer data)
{
  if(g_bAutoRotate)
    {
      g_xAngle+=5.0;
      g_yAngle+=7.0;
      g_zAngle+=3.0;
    }
  
  
  UpdateView();

  return False;		/* leave work proc active */
}

void resize(Widget w, XtPointer data, XtPointer callData)
{
    Dimension       width, height;

    if(made_current)
      {
        XtVaGetValues(w, XmNwidth, &width, XmNheight, &height, NULL);
        glViewport(0, 0, (GLint) width, (GLint) height);
        UpdateView();
      }
}

void quit(Widget w, XtPointer data, XtPointer callData)
{
    exit(0);
}

void toggleSpin()
{
  bInMotion = !bInMotion; /* toggle */
  
  if (bInMotion)
    workIdAnim = XtAppAddWorkProc(app, animate, NULL);
  else
    XtRemoveWorkProc(workIdAnim);
}


void map_state_changed(Widget w, XtPointer data, XEvent * event, Boolean * cont)
{  
  switch (event->type)
    {
    case MapNotify:
    if (bInMotion && workIdAnim != 0)
      workIdAnim = XtAppAddWorkProc(app, animate, NULL);
    break;
    case UnmapNotify:
      if (bInMotion) XtRemoveWorkProc(workIdAnim);
      break;
    }

}


void InitGLWindow(int argc, char** argv)
{

    XVisualInfo    *vi;
    Colormap        cmap;
    XSetWindowAttributes swa;
    GLXContext      cx;
    XEvent          event;
    GLboolean       needRedraw = GL_FALSE, recalcModelView = GL_TRUE;
    int		    dummy;

    /*Create window*/
    toplevel = XtAppInitialize(&app, "Life3D Viewer - OS A3 (v1.0)",
                               NULL, 0, &argc, argv,
			       fallbackResources, NULL, 0);

    dpy = XtDisplay(toplevel);      

    /*Find visual*/
    if (!(vi= glXChooseVisual(dpy, DefaultScreen(dpy), dblBuf)))
      {
        vi = glXChooseVisual(dpy, DefaultScreen(dpy), snglBuf);
        if (vi == NULL)
          XtAppError(app, "no RGB visual with depth buffer");
        doubleBuffer = GL_FALSE;
      }
    

    /* create an OpenGL rendering context */
    if(!(cx = glXCreateContext(dpy, vi,
                               None,   /* no sharing of display lists */
                               GL_TRUE /* direct rendering if possible */
                               )))
      {
        XtAppError(app,"life3dlib could not create an OpenGL rendering context");
      }
    

    XtAddEventHandler(toplevel, StructureNotifyMask, False,
                      map_state_changed, NULL);
    
    mainw = XmCreateMainWindow(toplevel, "mainw", NULL, 0);
    XtManageChild(mainw);

    /* create framed drawing area for OpenGL rendering */
    frame = XmCreateFrame(mainw, "frame", NULL, 0);
    XtManageChild(frame);

    /*Motif OpenGL drawing area*/
    glxarea = XtVaCreateManagedWidget
      ("glxarea", glwMDrawingAreaWidgetClass,
       frame, GLwNvisualInfo, vi, NULL);

    XtAddCallback(glxarea,
                  XmNexposeCallback,
                  (XtCallbackProc)draw, NULL);
    XtAddCallback(glxarea,
                  XmNresizeCallback,
                  resize, NULL);
    
    /* set up application's window layout */
    XmMainWindowSetAreas(mainw,
                         NULL,
                         NULL,
                         NULL, NULL, frame);
    XtRealizeWidget(toplevel);
    /*
     * Once widget is realized (ie, associated with a created X window), we
     * can bind the OpenGL rendering context to the window.
     */
    glXMakeCurrent(dpy, XtWindow(glxarea), cx);
    made_current = GL_TRUE;
}

void InitOpenGLState()
{
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glEnable(GL_LIGHTING);
  glEnable(GL_COLOR_MATERIAL);
  
  
  glDepthFunc(GL_LESS);    /* pedantic, GL_LESS is the default */
  glClearDepth(1.0);       /* pedantic, 1.0 is the default */
  /* frame buffer clears should be to black */
  glClearColor(0.0, 0.0, 0.0, 0.0);
  
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  makeCubeList();
  make2DGridList();
    
  /* set up projection transform */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-1.0, 1.0, -1.0, 1.0, 1.0, 10.0);
  /* establish initial viewport */
  glViewport(0, 0, 600, 300);
  
  
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
  glLightfv(GL_LIGHT0, GL_AMBIENT, lightZeroAmbient);
  glEnable(GL_LIGHT0);
}


/*Library interface functions*/

int LVLibInit(int argc,char **argv)
{
  InitGLWindow(argc,argv);
  
  /* setup OpenGL state */
  InitOpenGLState();

  bLibInitDone=TRUE;
  /*  XtAppMainLoop(app); */
  return 0;
}


int LVLibCleanup()
{
  /*nothing to clean*/
  return TRUE;
}

#define LVL_CUR_VERSION 1
int LVLLibGetVersion()
{
  return LVL_CUR_VERSION;
}

int LVLibGetNumCellsOnSide()
{
  return N;
}


int LVLibUpdateDisplay()
{
  if(!bLibInitDone)
    return FALSE;
  
  animate(NULL);
  return TRUE;
}

/*Set the cells for each face*/
int LVLibSetFaceCells(int faceN,int cells[][N][N])
{
  int i,j;
  if(faceN < 0 || faceN > 5)
    return FALSE;
  
  for(i=0;i<N;i++)
    for(j=0;j<N;j++)
      {        
        g_Faces[faceN][i][j]=cells[faceN][i][j];
      }
  

  return TRUE;
}

/*Starts cube rotating. Must be called at startup.
  Subsequent calls toggle rotation*/
void LVLibStartDisplay()
{
  toggleSpin();
  UpdateView();
}


/*Set the autorotation*/
/*if flag is True auto rotate enabled*/
/*On by default*/
extern void LVLibSetAutoRotate(LVLBool flag)
{
  g_bAutoRotate=flag;
}


/*Angle accessors*/
extern void LVLibSetXangle(float x)
{
  g_xAngle=(GLfloat)x;
}

extern void LVLibSetYangle(float y)
{
  g_yAngle=(GLfloat)y;
}

extern void LVLibSetZangle(float z)
{
  g_zAngle=(GLfloat)z;
}

extern float LVLibGetXangle()
{
  return (float)g_xAngle;
}

extern float LVLibGetYangle()
{
  return (float)g_yAngle;
}

extern float LVLibGetZangle()
{
  return (float)g_zAngle;
}
