/* lifedemo.c
 * Quick demo on how to use the life3d viewer library
 * Operating Systems - 304-427A, Fall 1998
 *
 * Nadim Abdo
 */

#include "stdio.h"
#include "life3dviewer.h"

/*N is the number of cells per edge if you change
 this you'll need to update CELLS_PER_EDGE in the viewer lib
 and recompile it*/
#define N 6
/*Cube faces*/
#define NUM_FACES 6

#define CELL_ALIVE      0x01
#define CELL_DEAD       0x00
/*bitmasks for next generation info*/
#define CELL_ALIVE_NEXT 0x02 /*next generation*/
#define CELL_DEAD_NEXT  0x01


/*Not particularly efficient implementation of life*/
/*You are more than welcome to implement your own for the
  assignment.*/
void RunLife(int face,int cells[][N][N])
{
  int i,j;
  int n;
  for(i=0;i<N;i++)
    {
      for(j=0;j<N;j++)
        {
          /*count neighbours naively*/
          n=0;
          if((i-1 >= 0) &&
             (CELL_ALIVE & cells[face][i-1][j]))
            n++;
          if((i+1 < N) &&
             (CELL_ALIVE & cells[face][i+1][j]))
            n++;
          if((j-1 >= 0) &&
             (CELL_ALIVE & cells[face][i][j-1]))
            n++;
          if((j+1 < N) &&
             (CELL_ALIVE & cells[face][i][j+1]))
            n++;
          /*diagonals*/
          if((i+1 < N) && (j+1 <N) &&
             (CELL_ALIVE & cells[face][i+1][j+1]))
            n++;
          if((i-1 >= 0) && (j-1 >=0) &&
             (CELL_ALIVE & cells[face][i-1][j-1]))
            n++;
          if((i+1 < N) && (j-1 >=0) &&
             (CELL_ALIVE & cells[face][i+1][j-1]))
            n++;
          if((i-1 >= 0) && (j+1 <N) &&
             (CELL_ALIVE & cells[face][i-1][j+1]))
            n++;
          
          
          /*A cell is alive on the next generation
            if
            a) it is dead and has 3 neighbours
            b) it is alive and has 2 or 3 neighbours
          */
          if( (3==n)  ||
              ((2==n) && (CELL_ALIVE & cells[face][i][j])))
            {
              cells[face][i][j]|=CELL_ALIVE_NEXT;

            }
          
          else
            cells[face][i][j]&=CELL_DEAD_NEXT;
        }
    }
  /*Advance to next generation*/
  for(i=0;i<N;i++)
    for(j=0;j<N;j++)
      {
        if(cells[face][i][j] & CELL_ALIVE_NEXT)
          cells[face][i][j]=CELL_ALIVE;
        else
          cells[face][i][j]=CELL_DEAD;
      }
}



int main(int argc,char** argv)
{
  int i;
  int n,j,k;

  int input_key;
  int faces[NUM_FACES][N][N];

  
  printf("Starting 3d life demo...\n");
  printf("Ctrl-c or kill the process to quit\n");

  /*clear all faces*/
  for(n=0;n<NUM_FACES;n++)
    for(j=0;j<N;j++)
      for(k=0;k<N;k++)
        faces[n][j][k]=CELL_DEAD;

  srand(getpid());
  
  for(n=0;n<4;n++)
    for(i=0;i<N;i++)
      for(j=0;j<N;j++)
        faces[n][i][j]=rand()%2;
  /*random initial patterns quickly stabilize
    on such a small grid*/

  /*cross*/
  for(i=0;i<N;i++)
    {
      faces[5][i][3]=CELL_ALIVE;
      faces[5][4][i]=CELL_ALIVE;
    }
  

  LVLibInit(argc,argv);
  LVLibStartDisplay();
  /*Enable auto rotation (spinning cube)
    set this to LVLFalse to disable automatic spinning
    then use the LVLSet?angle() calls to set whatever
    viewing angles you want. Do this before calling
    LVLibUpdateDisplay*/
  LVLibSetAutoRotate(LVLTrue);
  
  for(;;)
    {
      /*Run life on each of the faces*/
      /*Unlike the assignment, there is no border
        information exchange*/
      for(i=0;i<NUM_FACES;i++)
        {
          RunLife(i,faces);
          /*Update the life viewer with the new
            face information, i is the cell number
            and faces is the face array*/
          LVLibSetFaceCells(i,faces);
        }
      /*Update the display with the new face information*/
      LVLibUpdateDisplay();
    }
  return 0;
}
